package com.chelvc.framework.base.context;

import java.net.URI;
import java.util.Map;
import java.util.function.Function;

import com.chelvc.framework.common.util.ErrorUtils;
import lombok.NonNull;
import org.apache.http.NoHttpResponseException;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpMethod;
import org.springframework.http.RequestEntity;
import org.springframework.web.client.RestTemplate;

/**
 * Rest上下文工具类
 *
 * @author Woody
 * @date 2024/6/5
 */
public final class RestContextHolder {
    /**
     * RestTemplate实例
     */
    private static RestTemplate TEMPLATE;

    private RestContextHolder() {
    }

    /**
     * 获取RestTemplate实例
     *
     * @return RestTemplate实例
     */
    public static RestTemplate getTemplate() {
        if (TEMPLATE == null) {
            synchronized (RestTemplate.class) {
                if (TEMPLATE == null) {
                    TEMPLATE = ApplicationContextHolder.getBean(RestTemplate.class);
                }
            }
        }
        return TEMPLATE;
    }

    /**
     * 执行接口请求
     *
     * @param function 请求回调函数
     * @param <T>      结果类型
     * @return 请求结果
     */
    public static <T> T execute(@NonNull Function<RestTemplate, T> function) {
        try {
            return function.apply(getTemplate());
        } catch (Exception e) {
            if (ErrorUtils.contains(e, NoHttpResponseException.class)) {
                // 请求重试
                return function.apply(new RestTemplate());
            }
            throw e;
        }
    }

    /**
     * GET方式请求
     *
     * @param url  资源地址
     * @param type 对象类型
     * @param <T>  结果类型
     * @return 请求结果
     */
    public static <T> T get(@NonNull URI url, @NonNull Class<T> type) {
        return execute(rest -> rest.getForObject(url, type));
    }

    /**
     * GET方式请求
     *
     * @param url        资源地址
     * @param type       对象类型
     * @param parameters 请求参数
     * @param <T>        结果类型
     * @return 请求结果
     */
    public static <T> T get(@NonNull String url, @NonNull Class<T> type, @NonNull Object... parameters) {
        return execute(rest -> rest.getForObject(url, type, parameters));
    }

    /**
     * GET方式请求
     *
     * @param url        资源地址
     * @param type       对象类型
     * @param parameters 请求参数
     * @param <T>        结果类型
     * @return 请求结果
     */
    public static <T> T get(@NonNull String url, @NonNull Class<T> type, @NonNull Map<String, ?> parameters) {
        return execute(rest -> rest.getForObject(url, type, parameters));
    }

    /**
     * POST方式请求
     *
     * @param url     资源地址
     * @param request 请求实体
     * @param type    对象类型
     * @param <T>     结果类型
     * @return 请求结果
     */
    public static <T> T post(@NonNull URI url, Object request, @NonNull Class<T> type) {
        return execute(rest -> rest.postForObject(url, request, type));
    }

    /**
     * POST方式请求
     *
     * @param url        资源地址
     * @param request    请求实体
     * @param type       对象类型
     * @param parameters 请求参数
     * @param <T>        结果类型
     * @return 请求结果
     */
    public static <T> T post(@NonNull String url, Object request, @NonNull Class<T> type,
                             @NonNull Object... parameters) {
        return execute(rest -> rest.postForObject(url, request, type, parameters));
    }

    /**
     * POST方式请求
     *
     * @param url        资源地址
     * @param request    请求实体
     * @param type       对象类型
     * @param parameters 请求参数
     * @param <T>        结果类型
     * @return 请求结果
     */
    public static <T> T post(@NonNull String url, Object request, @NonNull Class<T> type,
                             @NonNull Map<String, ?> parameters) {
        return execute(rest -> rest.postForObject(url, request, type, parameters));
    }

    /**
     * 接口调用
     *
     * @param entity 请求实体
     * @param type   对象类型
     * @param <T>    结果类型
     * @return 请求结果
     */
    public static <T> T exchange(@NonNull RequestEntity<?> entity, @NonNull Class<T> type) {
        return execute(rest -> rest.exchange(entity, type).getBody());
    }

    /**
     * 接口调用
     *
     * @param entity 请求实体
     * @param type   对象类型
     * @param <T>    结果类型
     * @return 请求结果
     */
    public static <T> T exchange(@NonNull RequestEntity<?> entity, @NonNull ParameterizedTypeReference<T> type) {
        return execute(rest -> rest.exchange(entity, type).getBody());
    }

    /**
     * 接口调用
     *
     * @param url    资源地址
     * @param method 请求方式
     * @param entity 请求实体
     * @param type   对象类型
     * @param <T>    结果类型
     * @return 请求结果
     */
    public static <T> T exchange(@NonNull URI url, @NonNull HttpMethod method, HttpEntity<?> entity,
                                 @NonNull Class<T> type) {
        return execute(rest -> rest.exchange(url, method, entity, type).getBody());
    }

    /**
     * 接口调用
     *
     * @param url        资源地址
     * @param method     请求方式
     * @param entity     请求实体
     * @param type       对象类型
     * @param parameters 请求参数
     * @param <T>        结果类型
     * @return 请求结果
     */
    public static <T> T exchange(@NonNull String url, @NonNull HttpMethod method, HttpEntity<?> entity,
                                 @NonNull Class<T> type, @NonNull Object... parameters) {
        return execute(rest -> rest.exchange(url, method, entity, type, parameters).getBody());
    }

    /**
     * 接口调用
     *
     * @param url        资源地址
     * @param method     请求方式
     * @param entity     请求实体
     * @param type       对象类型
     * @param parameters 请求参数
     * @param <T>        结果类型
     * @return 请求结果
     */
    public static <T> T exchange(@NonNull String url, @NonNull HttpMethod method, HttpEntity<?> entity,
                                 @NonNull Class<T> type, @NonNull Map<String, ?> parameters) {
        return execute(rest -> rest.exchange(url, method, entity, type, parameters).getBody());
    }

    /**
     * 接口调用
     *
     * @param url    资源地址
     * @param method 请求方式
     * @param entity 请求实体
     * @param type   对象类型
     * @param <T>    结果类型
     * @return 请求结果
     */
    public static <T> T exchange(@NonNull URI url, @NonNull HttpMethod method, HttpEntity<?> entity,
                                 @NonNull ParameterizedTypeReference<T> type) {
        return execute(rest -> rest.exchange(url, method, entity, type).getBody());
    }

    /**
     * 接口调用
     *
     * @param url        资源地址
     * @param method     请求方式
     * @param entity     请求实体
     * @param type       对象类型
     * @param parameters 请求参数
     * @param <T>        结果类型
     * @return 请求结果
     */
    public static <T> T exchange(@NonNull String url, @NonNull HttpMethod method, HttpEntity<?> entity,
                                 @NonNull ParameterizedTypeReference<T> type, @NonNull Object... parameters) {
        return execute(rest -> rest.exchange(url, method, entity, type, parameters).getBody());
    }

    /**
     * 接口调用
     *
     * @param url        资源地址
     * @param method     请求方式
     * @param entity     请求实体
     * @param type       对象类型
     * @param parameters 请求参数
     * @param <T>        结果类型
     * @return 请求结果
     */
    public static <T> T exchange(@NonNull String url, @NonNull HttpMethod method, HttpEntity<?> entity,
                                 @NonNull ParameterizedTypeReference<T> type, @NonNull Map<String, ?> parameters) {
        return execute(rest -> rest.exchange(url, method, entity, type, parameters).getBody());
    }
}
