package com.chelvc.framework.database.support;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.function.Function;

import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.conditions.query.ChainQuery;
import com.baomidou.mybatisplus.extension.service.IService;
import com.chelvc.framework.common.model.Pagination;
import com.chelvc.framework.common.model.Paging;
import com.chelvc.framework.database.context.DatabaseContextHolder;
import lombok.NonNull;

/**
 * 业务增强操作接口
 *
 * @param <T> 数据模型类型
 * @author Woody
 * @date 2024/4/3
 */
public interface EnhanceService<T> extends IService<T>, EventService<T> {
    /**
     * 获取增强Mapper
     *
     * @return Mapper实例
     */
    default EnhanceMapper<T> getEnhanceMapper() {
        return (EnhanceMapper<T>) this.getBaseMapper();
    }

    /**
     * 创建或忽略实体
     *
     * @param entity 实体对象实例
     * @return true/false
     */
    default boolean createIgnore(@NonNull T entity) {
        return this.getEnhanceMapper().createIgnore(entity) > 0;
    }

    /**
     * 创建或修改实体
     *
     * @param entity 实体对象实例
     * @return true/false
     */
    default boolean createUpdate(@NonNull T entity) {
        return this.createUpdate(entity, Updates.all(this.getEntityClass()));
    }

    /**
     * 创建或修改实体
     *
     * @param entity 实体对象实例
     * @param update 数据更新对象实例
     * @return true/false
     */
    default boolean createUpdate(@NonNull T entity, @NonNull Updates.Update<T> update) {
        return this.getEnhanceMapper().createUpdate(entity, update) > 0;
    }

    /**
     * 批量创建或忽略实体
     *
     * @param entities 实体对象实例集合
     * @return true/false
     */
    default boolean batchCreateIgnore(@NonNull Collection<T> entities) {
        return !entities.isEmpty() && this.getEnhanceMapper().batchCreateIgnore(entities) > 0;
    }

    /**
     * 批量创建或修改实体
     *
     * @param entities 实体对象实例集合
     * @return true/false
     */
    default boolean batchCreateUpdate(@NonNull Collection<T> entities) {
        return this.batchCreateUpdate(entities, Updates.all(this.getEntityClass()));
    }

    /**
     * 批量创建或修改实体
     *
     * @param entities 实体对象实例集合
     * @param update   数据更新对象实例
     * @return true/false
     */
    default boolean batchCreateUpdate(@NonNull Collection<T> entities, @NonNull Updates.Update<T> update) {
        return !entities.isEmpty() && this.getEnhanceMapper().batchCreateUpdate(entities, update) > 0;
    }

    /**
     * 列表查询
     *
     * @param select 查询列
     * @return 对象列表
     */
    default List<T> select(@NonNull Selects.Select<T> select) {
        return this.select(select, this.lambdaQuery());
    }

    /**
     * 列表查询
     *
     * @param select 查询列
     * @param type   对象类型
     * @param <R>    数据类型
     * @return 对象列表
     */
    @SuppressWarnings("rawTypes")
    default <R> List<R> select(@NonNull Selects.Select<T> select, @NonNull Class<R> type) {
        return this.select(select, this.lambdaQuery(), type);
    }

    /**
     * 列表查询
     *
     * @param select 查询列
     * @param query  查询对象
     * @return 对象列表
     */
    default List<T> select(@NonNull Selects.Select<T> select, @NonNull ChainQuery<T> query) {
        return this.getEnhanceMapper().getObjects(select, query.getWrapper());
    }

    /**
     * 列表查询
     *
     * @param select 查询列
     * @param query  查询对象
     * @param type   对象类型
     * @param <R>    数据类型
     * @return 对象列表
     */
    @SuppressWarnings({"rawTypes", "unchecked"})
    default <R> List<R> select(@NonNull Selects.Select<T> select, @NonNull ChainQuery<T> query,
                               @NonNull Class<R> type) {
        if (type == this.getEntityClass()) {
            return (List<R>) this.select(select, query);
        }
        return DatabaseContextHolder.deserialize(this.getEnhanceMapper().getMaps(select, query.getWrapper()), type);
    }

    /**
     * 分页查询
     *
     * @param paging 查询分页信息
     * @return 对象分页信息
     */
    default Pagination<T> paging(@NonNull Paging paging) {
        return this.paging(this.lambdaQuery(), paging);
    }

    /**
     * 分页查询
     *
     * @param paging  查询分页信息
     * @param adapter 数据转换适配器
     * @param <R>     目标数据类型
     * @return 对象分页信息
     */
    default <R> Pagination<R> paging(@NonNull Paging paging, @NonNull Function<List<T>, List<R>> adapter) {
        return this.paging(this.lambdaQuery(), paging, adapter);
    }

    /**
     * 分页查询
     *
     * @param query  查询对象
     * @param paging 查询分页信息
     * @return 对象分页信息
     */
    default Pagination<T> paging(@NonNull ChainQuery<T> query, @NonNull Paging paging) {
        return DatabaseContextHolder.paging(paging, query::page);
    }

    /**
     * 分页查询
     *
     * @param query   查询对象
     * @param paging  查询分页信息
     * @param adapter 数据转换适配器
     * @param <R>     目标数据类型
     * @return 对象分页信息
     */
    default <R> Pagination<R> paging(@NonNull ChainQuery<T> query, @NonNull Paging paging,
                                     @NonNull Function<List<T>, List<R>> adapter) {
        return DatabaseContextHolder.paging(paging, query::page, adapter);
    }

    /**
     * 分页查询
     *
     * @param select 查询列
     * @param paging 查询分页信息
     * @return 对象分页信息
     */
    default Pagination<T> paging(@NonNull Selects.Select<T> select, @NonNull Paging paging) {
        return this.paging(select, this.lambdaQuery(), paging);
    }

    /**
     * 分页查询
     *
     * @param select 查询列
     * @param query  查询对象
     * @param paging 分页信息
     * @return 对象分页信息
     */
    default Pagination<T> paging(@NonNull Selects.Select<T> select, @NonNull ChainQuery<T> query,
                                 @NonNull Paging paging) {
        IPage<T> page = DatabaseContextHolder.page(paging);
        return DatabaseContextHolder.pagination(this.getEnhanceMapper().getPage(page, select, query.getWrapper()));
    }

    /**
     * 分页查询
     *
     * @param select 查询列
     * @param paging 查询分页信息
     * @param type   对象类型
     * @param <R>    数据类型
     * @return 对象分页信息
     */
    default <R> Pagination<R> paging(@NonNull Selects.Select<T> select, @NonNull Paging paging,
                                     @NonNull Class<R> type) {
        return this.paging(select, this.lambdaQuery(), paging, type);
    }

    /**
     * 分页查询
     *
     * @param select 查询列
     * @param query  查询对象
     * @param paging 分页信息
     * @param type   对象类型
     * @param <R>    数据类型
     * @return 对象分页信息
     */
    @SuppressWarnings({"rawTypes", "unchecked"})
    default <R> Pagination<R> paging(@NonNull Selects.Select<T> select, @NonNull ChainQuery<T> query,
                                     @NonNull Paging paging, @NonNull Class<R> type) {
        if (type == this.getEntityClass()) {
            return (Pagination<R>) this.paging(select, query, paging);
        }
        IPage<Map<String, Object>> page = DatabaseContextHolder.page(paging);
        return DatabaseContextHolder.pagination(
                this.getEnhanceMapper().getMapsPage(page, select, query.getWrapper()),
                maps -> DatabaseContextHolder.deserialize(maps, type)
        );
    }
}
