package com.chelvc.framework.database.support;

import java.io.Serializable;
import java.util.Collections;
import java.util.Map;
import java.util.Objects;

import com.baomidou.mybatisplus.core.toolkit.support.SFunction;
import com.chelvc.framework.common.util.AssertUtils;
import com.chelvc.framework.database.context.DatabaseContextHolder;
import com.google.common.collect.Maps;
import lombok.NonNull;

/**
 * 自定义查询字段工具类
 *
 * @author Woody
 * @date 2024/5/27
 */
public final class Selects {
    private Selects() {
    }

    /**
     * 获取字段查询对象
     *
     * @param <T> 数据类型
     * @return 字段查询对象实例
     */
    public static <T> Select<T> select() {
        return new Select<>();
    }

    /**
     * 普通查询
     *
     * @param columns 列名数组
     * @param <T>     数据类型
     * @return 字段查询对象实例
     */
    public static <T> Select<T> of(@NonNull String... columns) {
        Select<T> select = select();
        return select.of(columns);
    }

    /**
     * 普通查询
     *
     * @param getters Getter方法数组
     * @param <T>     数据类型
     * @return 字段查询对象实例
     */
    @SafeVarargs
    public static <T> Select<T> of(@NonNull SFunction<T, ?>... getters) {
        Select<T> select = select();
        return select.of(getters);
    }

    /**
     * 普通查询
     *
     * @param column 列名
     * @param alias  别名
     * @param <T>    数据类型
     * @return 字段查询对象实例
     */
    public static <T> Select<T> of(@NonNull String column, @NonNull String alias) {
        Select<T> select = select();
        return select.of(column, alias);
    }

    /**
     * 普通查询
     *
     * @param getter Getter方法
     * @param alias  别名
     * @param <T>    数据类型
     * @return 字段查询对象实例
     */
    public static <T> Select<T> of(@NonNull SFunction<T, ?> getter, @NonNull String alias) {
        Select<T> select = select();
        return select.of(getter, alias);
    }

    /**
     * 求平均值
     *
     * @param column 列名
     * @param alias  别名
     * @param <T>    数据类型
     * @return 字段查询对象实例
     */
    public static <T> Select<T> avg(@NonNull String column, @NonNull String alias) {
        Select<T> select = select();
        return select.avg(column, alias);
    }

    /**
     * 求平均值
     *
     * @param getter Getter方法
     * @param alias  别名
     * @param <T>    数据类型
     * @return 字段查询对象实例
     */
    public static <T> Select<T> avg(@NonNull SFunction<T, ?> getter, @NonNull String alias) {
        Select<T> select = select();
        return select.avg(getter, alias);
    }

    /**
     * 求和
     *
     * @param column 列名
     * @param alias  别名
     * @param <T>    数据类型
     * @return 字段查询对象实例
     */
    public static <T> Select<T> sum(@NonNull String column, @NonNull String alias) {
        Select<T> select = select();
        return select.sum(column, alias);
    }

    /**
     * 求和
     *
     * @param getter Getter方法
     * @param alias  别名
     * @param <T>    数据类型
     * @return 字段查询对象实例
     */
    public static <T> Select<T> sum(@NonNull SFunction<T, ?> getter, @NonNull String alias) {
        Select<T> select = select();
        return select.sum(getter, alias);
    }

    /**
     * 求数量
     *
     * @param column 列名
     * @param alias  别名
     * @param <T>    数据类型
     * @return 字段查询对象实例
     */
    public static <T> Select<T> count(@NonNull String column, @NonNull String alias) {
        Select<T> select = select();
        return select.count(column, alias);
    }

    /**
     * 求数量
     *
     * @param column   列名
     * @param alias    别名
     * @param distinct 是否去重
     * @param <T>      数据类型
     * @return 字段查询对象实例
     */
    public static <T> Select<T> count(@NonNull String column, @NonNull String alias, boolean distinct) {
        Select<T> select = select();
        return select.count(column, alias, distinct);
    }

    /**
     * 求数量
     *
     * @param getter Getter方法
     * @param alias  别名
     * @param <T>    数据类型
     * @return 字段查询对象实例
     */
    public static <T> Select<T> count(@NonNull SFunction<T, ?> getter, @NonNull String alias) {
        Select<T> select = select();
        return select.count(getter, alias);
    }

    /**
     * 求数量
     *
     * @param getter   Getter方法
     * @param alias    别名
     * @param distinct 是否去重
     * @param <T>      数据类型
     * @return 字段查询对象实例
     */
    public static <T> Select<T> count(@NonNull SFunction<T, ?> getter, @NonNull String alias, boolean distinct) {
        Select<T> select = select();
        return select.count(getter, alias, distinct);
    }

    /**
     * 设置查询表达式
     *
     * @param alias      别名
     * @param expression 值表达式
     * @param <T>        数据类型
     * @return 字段查询对象实例
     */
    public static <T> Select<T> set(@NonNull String alias, @NonNull String expression) {
        Select<T> select = select();
        return select.set(alias, expression);
    }

    /**
     * 字段查询对象
     *
     * @param <T> 数据类型
     */
    public static class Select<T> implements Serializable {
        private final Map<String, String> expressions = Maps.newLinkedHashMap();

        /**
         * 普通查询
         *
         * @param columns 列名数组
         * @return 字段查询对象实例
         */
        public final Select<T> of(@NonNull String... columns) {
            for (String column : columns) {
                this.of(column, column);
            }
            return this;
        }

        /**
         * 普通查询
         *
         * @param getters Getter方法数组
         * @return 字段查询对象实例
         */
        @SafeVarargs
        public final Select<T> of(@NonNull SFunction<T, ?>... getters) {
            for (SFunction<T, ?> getter : getters) {
                this.of(DatabaseContextHolder.getter2column(getter));
            }
            return this;
        }

        /**
         * 普通查询
         *
         * @param column 列名
         * @param alias  别名
         * @return 字段查询对象实例
         */
        public final Select<T> of(@NonNull String column, @NonNull String alias) {
            return this.set(alias, column);
        }

        /**
         * 普通查询
         *
         * @param getter Getter方法
         * @param alias  别名
         * @return 字段查询对象实例
         */
        public final Select<T> of(@NonNull SFunction<T, ?> getter, @NonNull String alias) {
            return this.of(DatabaseContextHolder.getter2column(getter), alias);
        }

        /**
         * 求最小值
         *
         * @param column 列名
         * @param alias  别名
         * @return 字段查询对象实例
         */
        public final Select<T> min(@NonNull String column, @NonNull String alias) {
            return this.set(alias, "MIN(" + column + ")");
        }

        /**
         * 求最小值
         *
         * @param getter Getter方法
         * @param alias  别名
         * @return 字段查询对象实例
         */
        public final Select<T> min(@NonNull SFunction<T, ?> getter, @NonNull String alias) {
            return this.min(DatabaseContextHolder.getter2column(getter), alias);
        }

        /**
         * 求最大值
         *
         * @param column 列名
         * @param alias  别名
         * @return 字段查询对象实例
         */
        public final Select<T> max(@NonNull String column, @NonNull String alias) {
            return this.set(alias, "MAX(" + column + ")");
        }

        /**
         * 求最大值
         *
         * @param getter Getter方法
         * @param alias  别名
         * @return 字段查询对象实例
         */
        public final Select<T> max(@NonNull SFunction<T, ?> getter, @NonNull String alias) {
            return this.max(DatabaseContextHolder.getter2column(getter), alias);
        }

        /**
         * 求平均值
         *
         * @param column 列名
         * @param alias  别名
         * @return 字段查询对象实例
         */
        public final Select<T> avg(@NonNull String column, @NonNull String alias) {
            return this.set(alias, "AVG(" + column + ")");
        }

        /**
         * 求平均值
         *
         * @param getter Getter方法
         * @param alias  别名
         * @return 字段查询对象实例
         */
        public final Select<T> avg(@NonNull SFunction<T, ?> getter, @NonNull String alias) {
            return this.avg(DatabaseContextHolder.getter2column(getter), alias);
        }

        /**
         * 求和
         *
         * @param column 列名
         * @param alias  别名
         * @return 字段查询对象实例
         */
        public final Select<T> sum(@NonNull String column, @NonNull String alias) {
            return this.set(alias, "SUM(" + column + ")");
        }

        /**
         * 求和
         *
         * @param getter Getter方法
         * @param alias  别名
         * @return 字段查询对象实例
         */
        public final Select<T> sum(@NonNull SFunction<T, ?> getter, @NonNull String alias) {
            return this.sum(DatabaseContextHolder.getter2column(getter), alias);
        }

        /**
         * 求数量
         *
         * @param column 列名
         * @param alias  别名
         * @return 字段查询对象实例
         */
        public final Select<T> count(@NonNull String column, @NonNull String alias) {
            return this.count(column, alias, false);
        }

        /**
         * 求数量
         *
         * @param column   列名
         * @param alias    别名
         * @param distinct 是否去重
         * @return 字段查询对象实例
         */
        public final Select<T> count(@NonNull String column, @NonNull String alias, boolean distinct) {
            return this.set(alias, distinct ? "COUNT(DISTINCT " + column + ")" : "COUNT(" + column + ")");
        }

        /**
         * 求数量
         *
         * @param getter Getter方法
         * @param alias  别名
         * @return 字段查询对象实例
         */
        public final Select<T> count(@NonNull SFunction<T, ?> getter, @NonNull String alias) {
            return this.count(getter, alias, false);
        }

        /**
         * 求数量
         *
         * @param getter   Getter方法
         * @param alias    别名
         * @param distinct 是否去重
         * @return 字段查询对象实例
         */
        public final Select<T> count(@NonNull SFunction<T, ?> getter, @NonNull String alias, boolean distinct) {
            return this.count(DatabaseContextHolder.getter2column(getter), alias, distinct);
        }

        /**
         * 设置查询表达式
         *
         * @param alias      别名
         * @param expression 值表达式
         * @return 字段查询对象实例
         */
        public final Select<T> set(@NonNull String alias, @NonNull String expression) {
            String exists = this.expressions.put(alias, expression);
            AssertUtils.check(Objects.isNull(exists), () -> "Duplicate column alias: " + alias);
            return this;
        }

        /**
         * 获取查询字段别名/值表达式映射表
         *
         * @return 字段别名/值表达式映射表
         */
        public final Map<String, String> getExpressions() {
            return Collections.unmodifiableMap(this.expressions);
        }

        @Override
        public String toString() {
            return this.expressions.toString();
        }
    }
}
