package com.chelvc.framework.wechat;

import java.time.Duration;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.UUID;
import java.util.concurrent.TimeUnit;
import java.util.function.Supplier;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.context.RestContextHolder;
import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.common.util.AssertUtils;
import com.chelvc.framework.common.util.DecimalUtils;
import com.chelvc.framework.common.util.HostUtils;
import com.chelvc.framework.common.util.ObjectUtils;
import com.chelvc.framework.common.util.StringUtils;
import com.chelvc.framework.redis.context.RedisContextHolder;
import com.chelvc.framework.wechat.config.WechatProperties;
import com.chelvc.framework.wechat.context.WechatContextHolder;
import com.github.wxpay.sdk.WXPayUtil;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;

/**
 * 微信处理默认实现
 *
 * @author Woody
 * @date 2024/1/30
 */
@Component
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class DefaultWechatHandler implements WechatHandler {
    /**
     * Scheme过期时间（天）
     */
    private static final int SCHEME_EXPIRE_INTERVAL = 30;

    /**
     * 访问令牌标识前缀
     */
    private static final String ACCESS_TOKEN_PREFIX = "wechat.token.";

    /**
     * 获取用户信息接口地址
     */
    private static final String USER_URL =
            "https://api.weixin.qq.com/cgi-bin/user/info?access_token=%s&openid=%s&lang=zh_CN";

    /**
     * 微信下单接口地址
     */
    private static final String UNIFIEDORDER_URL = "https://api.mch.weixin.qq.com/pay/unifiedorder";

    /**
     * 微信获取手机号接口地址
     */
    private static final String CODE2MOBILE_URL =
            "https://api.weixin.qq.com/wxa/business/getuserphonenumber?access_token=";

    /**
     * 微信登录接口地址
     */
    private static final String JSCODE2SESSION_URL =
            "https://api.weixin.qq.com/sns/jscode2session?appid=%s&secret=%s&js_code=%s&grant_type=authorization_code";

    /**
     * 获取访问令牌接口地址
     */
    private static final String ACCESS_TOKEN_URL =
            "https://api.weixin.qq.com/cgi-bin/token?grant_type=client_credential&appid=%s&secret=%s";

    /**
     * 获取小程序schemeCode接口地址
     */
    private static final String GENERATE_SCHEME_URL = "https://api.weixin.qq.com/wxa/generatescheme?access_token=";

    private final WechatProperties properties;
    private final Map<String, String> tokens = Maps.newConcurrentMap();

    /**
     * 查找支付配置
     *
     * @param type 支付类型
     * @return 支付配置
     */
    private WechatProperties.Payment lookupPayment(String type) {
        List<WechatProperties.Payment> payments = this.properties.getPayments();
        WechatProperties.Payment payment = CollectionUtils.isEmpty(payments) ? null :
                payments.stream().filter(config -> Objects.equals(config.getType(), type)).findFirst().orElse(null);
        return AssertUtils.nonnull(payment, () -> "Not support payment type: " + type);
    }

    /**
     * 参数签名
     *
     * @param payment    支付配置
     * @param parameters 签名参数
     * @return 签名信息
     */
    private String sign(WechatProperties.Payment payment, Map<String, String> parameters) {
        return WechatContextHolder.sign(Objects.requireNonNull(payment.getMchkey()), parameters);
    }

    /**
     * 构建支付参数
     *
     * @param payment 支付配置
     * @param request 支付请求参数
     * @return 支付参数键/值对
     */
    private Map<String, String> buildPaymentParameter(WechatProperties.Payment payment, WechatPayRequest request) {
        String mchid = Objects.requireNonNull(payment.getMchid());
        String callback = Objects.requireNonNull(payment.getCallback());
        Map<String, String> parameters = Maps.newHashMap();
        parameters.put("appid", payment.getAppid());
        parameters.put("mch_id", mchid);
        parameters.put("body", request.getComment());
        parameters.put("out_trade_no", request.getOrder());
        if (StringUtils.notEmpty(request.getContext())) {
            parameters.put("attach", request.getContext());
        }
        parameters.put("fee_type", "CNY");
        parameters.put("total_fee", String.valueOf(request.getAmount().multiply(DecimalUtils.HUNDRED).intValue()));
        parameters.put("spbill_create_ip", HostUtils.LOCAL_ADDRESS);
        parameters.put("notify_url", callback);
        parameters.put("trade_type", payment.getMode().name());
        parameters.put("nonce_str", WXPayUtil.generateNonceStr());
        return parameters;
    }

    @Override
    public String getAppid() {
        return this.properties.getAppid();
    }

    @Override
    public String getAccessToken() {
        return this.getAccessToken(this.properties.getAppid(), this.properties.getSecret());
    }

    @Override
    public String refreshAccessToken() {
        return this.refreshAccessToken(this.properties.getAppid(), this.properties.getSecret());
    }

    @Override
    public String getAccessToken(@NonNull String appid, @NonNull String secret) {
        return this.tokens.computeIfAbsent(ACCESS_TOKEN_PREFIX + appid, k -> this.refreshAccessToken(appid, secret));
    }

    @Override
    public String refreshAccessToken(@NonNull String appid, @NonNull String secret) {
        String key = ACCESS_TOKEN_PREFIX + appid;
        String token = ApplicationContextHolder.getProperty(key);
        if (StringUtils.notEmpty(token)) {
            return token;
        }
        return RedisContextHolder.lockAround(key + ":lock", () -> {
            String value = (String) RedisContextHolder.getDefaultTemplate().opsForValue().get(key);
            if (StringUtils.isEmpty(value) || Objects.equals(value, this.tokens.get(key))) {
                String url = String.format(ACCESS_TOKEN_URL, appid, secret);
                WechatAccessToken accessToken = RestContextHolder.get(url, WechatAccessToken.class);
                value = accessToken.getToken();
                long duration = accessToken.getDuration() - 60;
                RedisContextHolder.getDefaultTemplate().opsForValue().set(key, value, duration, TimeUnit.SECONDS);
                this.tokens.put(key, value);
            }
            return value;
        }, () -> {
            throw new IllegalStateException("Get wechat access token timeout");
        });
    }

    @Override
    public String getScheme() {
        return this.getScheme(null, null, null);
    }

    @Override
    public String getScheme(String path, String query, String environment) {
        // 尝试从Redis中获取Scheme
        path = StringUtils.ifEmpty(path, StringUtils.EMPTY);
        query = StringUtils.ifEmpty(query, StringUtils.EMPTY);
        environment = StringUtils.ifEmpty(environment, StringUtils.EMPTY);
        HttpSession session = ObjectUtils.ifNull(SessionContextHolder.getRequest(), HttpServletRequest::getSession);
        String id = ObjectUtils.ifNull(
                ObjectUtils.ifNull(session, HttpSession::getId), () -> UUID.randomUUID().toString()
        );
        String key = "wechat:scheme:" + id + ":" + environment + ":" + path + ":" + query;
        String scheme = (String) RedisContextHolder.getDefaultTemplate().opsForValue().get(key);
        if (StringUtils.notEmpty(scheme)) {
            return scheme;
        }

        // 重新获取Scheme
        Map<String, Object> body = ImmutableMap.of(
                "jump_wxa", ImmutableMap.of(
                        "path", path,
                        "query", query,
                        "env_version", environment
                ),
                "expire_type", 1,
                "expire_interval", SCHEME_EXPIRE_INTERVAL
        );
        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_JSON);
        HttpEntity<?> entity = new HttpEntity<>(body, headers);
        scheme = WechatContextHolder.exchange(
                init -> GENERATE_SCHEME_URL + (init ? this.getAccessToken() : this.refreshAccessToken()),
                HttpMethod.POST, entity, WechatScheme.class
        ).getOpenlink();
        RedisContextHolder.getDefaultTemplate().opsForValue()
                .set(key, scheme, Duration.ofDays(SCHEME_EXPIRE_INTERVAL - 1));
        return scheme;
    }

    @Override
    public WechatUser getUser(@NonNull String accessToken, @NonNull String openid) {
        return this.getUser(accessToken, openid, this::refreshAccessToken);
    }

    @Override
    public WechatUser getUser(@NonNull String accessToken, @NonNull String openid,
                              @NonNull Supplier<String> tokenRefresher) {
        return WechatContextHolder.get(
                init -> String.format(USER_URL, init ? accessToken : tokenRefresher.get(), openid),
                WechatUser.class
        );
    }

    @Override
    public String code2mobile(@NonNull String code) {
        WechatMobile.Mobile mobile = WechatContextHolder.post(
                init -> CODE2MOBILE_URL + (init ? this.getAccessToken() : this.refreshAccessToken()),
                ImmutableMap.of("code", code), WechatMobile.class
        ).getMobile();
        return Objects.requireNonNull(mobile).getNumber();
    }

    @Override
    public WechatSession code2session(@NonNull String code) {
        String url = String.format(JSCODE2SESSION_URL, this.properties.getAppid(), this.properties.getSecret(), code);
        return RestContextHolder.get(url, WechatSession.class);
    }

    @Override
    public String sign(@NonNull String type, @NonNull Map<String, String> parameters) {
        return this.sign(this.lookupPayment(type), parameters);
    }

    @Override
    public boolean validate(@NonNull String type, @NonNull Map<String, String> parameters) {
        if (ObjectUtils.isEmpty(this.properties.getPayments())) {
            return false;
        }
        String sign = parameters.get("sign");
        if (StringUtils.isEmpty(sign)) {
            return false;
        }
        for (WechatProperties.Payment payment : this.properties.getPayments()) {
            if (Objects.equals(payment.getType(), type) && Objects.equals(this.sign(payment, parameters), sign)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public WechatUnifiedOrder unifiedorder(@NonNull String type, @NonNull WechatPayRequest request) {
        // 获取支付配置
        WechatProperties.Payment payment = this.lookupPayment(type);

        // 构建请求参数
        Map<String, String> parameters = this.buildPaymentParameter(payment, request);

        if (payment.getMode() == PayMode.MWEB) {
            // 如果是H5支付则需要指定scene_info，场景信息
            parameters.put("scene_info", Objects.requireNonNull(request.getScene()));
        } else if (payment.getMode() == PayMode.JSAPI) {
            // 如果是小程序支付则需要指定openid
            parameters.put("openid", Objects.requireNonNull(request.getOpenid()));
        }

        // 请求参数签名
        parameters.put("sign", this.sign(payment, parameters));

        // 将请求参数转换成xml请求体
        String body = WechatContextHolder.map2xml(parameters);

        // 发起下单请求
        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.TEXT_XML);
        HttpEntity<?> entity = new HttpEntity<>(body, headers);
        String response = RestContextHolder.execute(rest -> rest.exchange(
                UNIFIEDORDER_URL, HttpMethod.POST, entity, String.class
        )).getBody();

        // 验证相应结果
        Map<String, String> result = ObjectUtils.ifNull(response, WechatContextHolder::xml2map);
        if (StringUtils.isEmpty(response) || !Objects.equals(result.get("return_code"), "SUCCESS")
                || !Objects.equals(result.get("result_code"), "SUCCESS")
                || !Objects.equals(result.remove("sign"), this.sign(payment, result))) {
            throw new RuntimeException("Wechat unifiedorder failed");
        }

        // 构建微信统一下单信息
        return WechatUnifiedOrder.builder().mode(payment.getMode()).appid(result.get("appid"))
                .mchid(result.get("mch_id")).nonce(result.get("nonce_str")).prepayid(result.get("prepay_id"))
                .qrcode(result.get("code_url")).redirect(result.get("mweb_url")).build();
    }
}
