package com.chelvc.framework.wechat;

import java.security.MessageDigest;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.function.BiConsumer;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.crypto.Cipher;

import com.chelvc.framework.base.context.RestContextHolder;
import com.chelvc.framework.common.util.AESUtils;
import com.chelvc.framework.common.util.AssertUtils;
import com.chelvc.framework.common.util.ObjectUtils;
import com.chelvc.framework.common.util.StringUtils;
import com.chelvc.framework.wechat.config.WechatProperties;
import com.chelvc.framework.wechat.context.WechatContextHolder;
import com.google.common.collect.ImmutableMap;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.util.CollectionUtils;

/**
 * 微信公众号操作默认实现
 *
 * @author Woody
 * @date 2024/1/30
 */
@Slf4j
@RequiredArgsConstructor(onConstructor = @__(@Autowired))
public class DefaultWechatPublicHandler implements WechatPublicHandler {
    /**
     * 发送微信服务号模板消息接口地址
     */
    private static final String SEND_TEMPLATE_MESSAGE_URL =
            "https://api.weixin.qq.com/cgi-bin/message/template/send?access_token=";

    /**
     * 批量获取关注用户openid地址
     */
    private static final String OPENID_URL =
            "https://api.weixin.qq.com/cgi-bin/user/get?access_token=%s&next_openid=%s";

    /**
     * 获取公众号粉丝unionId信息地址
     */
    private static final String USER_URL = "https://api.weixin.qq.com/cgi-bin/user/info/batchget?access_token=";

    /**
     * 获取微信网页令牌接口地址
     */
    private static final String CODE2TOKEN_URL = "https://api.weixin.qq.com/sns/oauth2/access_token" +
            "?appid=%s&secret=%s&code=%s&grant_type=authorization_code";

    private final WechatHandler wechatHandler;
    private final WechatProperties.Public properties;

    @Override
    public String getId() {
        return StringUtils.ifEmpty(this.properties.getId(), this.properties.getAppid());
    }

    @Override
    public String getAppid() {
        return this.properties.getAppid();
    }

    @Override
    public String getAccessToken() {
        return this.wechatHandler.getAccessToken(this.properties.getAppid(), this.properties.getSecret());
    }

    @Override
    public String refreshAccessToken() {
        return this.wechatHandler.refreshAccessToken(this.properties.getAppid(), this.properties.getSecret());
    }

    @Override
    public WechatHandler getWechatHandler() {
        return this.wechatHandler;
    }

    @Override
    public WechatWebToken code2token(@NonNull String code) {
        String url = String.format(CODE2TOKEN_URL, this.properties.getAppid(), this.properties.getSecret(), code);
        return RestContextHolder.get(url, WechatWebToken.class);
    }

    @Override
    public String decrypt(String ciphertext) {
        if (StringUtils.isEmpty(ciphertext)) {
            return ciphertext;
        }
        String secret = AssertUtils.nonempty(this.properties.getKey(), () -> "Secret is missing");
        Cipher cipher = AESUtils.lookupDecrypter(AESUtils.CBC_NOPADDING + secret, () -> {
            Pair<byte[], byte[]> pair = AESUtils.getDefaultCBCCertificate(secret);
            return AESUtils.getCipher(AESUtils.CBC_NOPADDING, Cipher.DECRYPT_MODE, pair.getLeft(), pair.getRight());
        });
        return WechatContextHolder.decrypt(cipher, ciphertext);
    }

    @Override
    public WechatUser getUser(@NonNull String openid) {
        return this.wechatHandler.getUser(this.getAccessToken(), openid, this::refreshAccessToken);
    }

    @Override
    public void iterateOpenids(@NonNull Consumer<List<String>> consumer) {
        this.iterateOpenids((openids, total) -> consumer.accept(openids));
    }

    @Override
    public void iterateOpenids(@NonNull BiConsumer<List<String>, Integer> consumer) {
        int total = 0;
        String next = null;
        do {
            final String _next = ObjectUtils.ifNull(next, StringUtils.EMPTY);
            WechatPublicOpenids response = WechatContextHolder.get(
                    init -> String.format(OPENID_URL, init ? this.getAccessToken() : this.refreshAccessToken(), _next),
                    WechatPublicOpenids.class
            );
            List<String> openids = ObjectUtils.ifNull(response.getData(), WechatPublicOpenids.Openids::getOpenid);
            if (CollectionUtils.isEmpty(openids)) {
                break;
            }
            consumer.accept(openids, total += openids.size());
            next = response.getNextOpenid();
        } while (StringUtils.notEmpty(next));
    }

    @Override
    public List<WechatPublicSubscriber> getSubscribers(@NonNull Collection<String> openids) {
        if (CollectionUtils.isEmpty(openids)) {
            return Collections.emptyList();
        }
        Map<?, ?> body = ImmutableMap.of(
                "user_list",
                openids.stream().map(openid -> ImmutableMap.of("openid", openid)).collect(Collectors.toList())
        );
        WechatPublicSubscribers response = WechatContextHolder.post(
                init -> USER_URL + (init ? this.getAccessToken() : this.refreshAccessToken()),
                body, WechatPublicSubscribers.class
        );
        return ObjectUtils.ifNull(response.getSubscribers(), Collections::emptyList);
    }

    @Override
    public String sign(long timestamp, @NonNull String... elements) {
        MessageDigest digest = DigestUtils.getDigest("SHA-1");
        String original = Stream.concat(
                Stream.of(this.properties.getToken(), String.valueOf(timestamp)), Stream.of(elements)
        ).sorted().collect(Collectors.joining());
        return Hex.encodeHexString(digest.digest(original.getBytes()));
    }

    @Override
    public long push(@NonNull String openid, @NonNull String template, @NonNull Map<?, ?> parameters) {
        return this.push(openid, template, StringUtils.EMPTY, parameters);
    }

    @Override
    public long push(@NonNull String openid, @NonNull String template, @NonNull String redirect,
                     @NonNull Map<?, ?> parameters) {
        boolean debug = log.isDebugEnabled();
        if (debug) {
            log.debug("Wechat public push request: {}, {}, {}", openid, template, parameters);
        }
        String appid = (redirect = redirect.trim()).isEmpty() ? StringUtils.EMPTY : this.wechatHandler.getAppid();
        Map<String, Object> body = ImmutableMap.of(
                "touser", openid,
                "template_id", template,
                "url", StringUtils.EMPTY,
                "miniprogram", ImmutableMap.of("appid", appid, "pagepath", redirect),
                "data", parameters
        );
        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_JSON);
        HttpEntity<?> entity = new HttpEntity<>(body, headers);
        WechatPublicMessage response;
        try {
            response = WechatContextHolder.exchange(
                    init -> SEND_TEMPLATE_MESSAGE_URL + (init ? this.getAccessToken() : this.refreshAccessToken()),
                    HttpMethod.POST, entity, WechatPublicMessage.class
            );
        } catch (Exception e) {
            log.warn("Wechat public push failed: {}, {}, {}", openid, template, e.getMessage());
            return 0;
        }
        if (debug) {
            log.debug("Wechat public push response: {}, {}, {}", openid, template, response);
        }
        return response.getId();
    }
}
